import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;

import org.json.JSONException;
import org.json.JSONObject;

public class bitNETs {
	
	static final String BITNETS = "bitNETs";
	
	static final String AUTHOR = "Lovro Subelj (lovro.subelj@fri.uni-lj.si)";
	
	static final boolean PRINT = true;
	
	static final double GAP = 0.125;
	
	static final String TAB = "   ";
	
	static String PATH = ".";
	
	static String FILE = "bitcoin";
	
	static boolean MAXIMUM = true;
	
	static double CORRELATION = 0.95;
	
	static int SEGMENT = 15;
	
	static int QUANTILE = 1000;
	
	static int INSTANCES = 0;
	
	/*
	 * 		MAIN
	 */

	public static void main(String[] args) {
		long tic = System.currentTimeMillis();
		
		JSONObject JSON; String json = "";
		
		try {
			BufferedReader reader = new BufferedReader(new FileReader(BITNETS + ".json"));

			String line;
			while ((line = reader.readLine()) != null)
				json += " " + line.trim();

			reader.close();
		} catch (IOException e) {

		}

		try {
			JSON = new JSONObject(json);
		} catch (JSONException e) {
			JSON = new JSONObject();
		}
		
		String path = PATH;
		try {
			if (JSON.has("path"))
				path = JSON.getString("path");
		} catch (JSONException e) {
			
		}
		
		String file = FILE;
		try {
			if (JSON.has("file"))
				file = JSON.getString("file");
		} catch (JSONException e) {
			
		}

		boolean maximum = MAXIMUM;
		try {
			if (JSON.has("maximum"))
				maximum = JSON.getBoolean("maximum");
		} catch (JSONException e) {
			
		}
		
		double correlation = CORRELATION;
		try {
			if (JSON.has("correlation"))
				correlation = JSON.getDouble("correlation");
		} catch (JSONException e) {
			
		}
		
		int segment = SEGMENT;
		try {
			if (JSON.has("segment"))
				segment = JSON.getInt("segment");
		} catch (JSONException e) {
			
		}
		
		int quantile = QUANTILE;
		try {
			if (JSON.has("quantile"))
				quantile = JSON.getInt("quantile");
		} catch (JSONException e) {
			
		}
		
		int instances = INSTANCES;
		try {
			if (JSON.has("instances"))
				instances = JSON.getInt("instances");
		} catch (JSONException e) {
			
		}
		
		print("\n");
		
		Series series;
		try {
			print(BITNETS + " - Parsing time series in JSON format from \"" + file + ".json\"");
			series = json(path, file, maximum, segment, quantile, correlation); ok();
		} catch (IOException e) {
			error(); e.printStackTrace();
			
			return;
		}
		
		print("\n");
		
		try {
			print(BITNETS + " - Writing time series in ARFF format to \"" + file + ".arff\"");
			arff(path, file, series); ok();
			
			print(BITNETS + " - Writing time series in TAB format to \"" + file + ".tab\"");
			tab(path, file, series); ok();
		} catch (IOException e) {
			error(); e.printStackTrace();
		}
		
		print("\n");
		
		for (Type type: new Type[] {Type.CYCLE, Type.CORRELATION, Type.VISIBILITY, Type.TRANSITION})
			try {
				print(BITNETS + " - Generating " + type.toString()  + " from time series");
				Network network = network(type, series); ok();
				
				print(BITNETS + " - Writing " + network.name()  + " in bitNET format to \"" + file + "." + network.format() + "\"");
				bitnet(path, file, network); ok();
				
				print("\n");
			} catch (IOException e) {
				error(); e.printStackTrace();
			}
		
		if (instances > 0)
			try {
				print(BITNETS + " - Writing random time series in JSON format to \"" + file + "-" + instances + ".json\"");
				json(path, file + "-" + instances, instances); ok();
				
				print("\n");
			} catch (IOException e) {
				error(); e.printStackTrace();
			}
		
		time(tic);
	}
	
	/*
	 * 		NETWORKS
	 */
	
	static Network network(Type type, Series series) {
		switch (type) {
		case CYCLE:
			return cycle(series);
		case CORRELATION:
			return correlation(series);
		case VISIBILITY:
			return visibility(series);
		case TRANSITION:
			return transition(series);
		default:
			break;
		}
		
		return null;
	}
	
	static CycleNetwork cycle(Series series) {
		print("\n\t0%");
		
		long gap = (long)Math.ceil(GAP * series.getCycles().size() * (series.getCycles().size() + 1.0) / 2.0); int index = 0;
		
		Map<Cycle, Node> nodes = new HashMap<Cycle, Node>();
		for (Cycle cycle: series.getCycles()) {
			if (index > 0 && index % gap == 0) print(TAB + Math.round(200.0 * index / (series.getCycles().size() * (series.getCycles().size() + 1.0))) + "%"); index++;
			
			nodes.put(cycle, new Node(cycle.getId(), cycle.toString()));
		}
		
		List<Link> links = new ArrayList<Link>();
		for (int i = 0; i < series.getCycles().size(); i++)
			for (int j = i + 1; j < series.getCycles().size(); j++) {
				if (index > 0 && index % gap == 0) print(TAB + Math.round(200.0 * index / (series.getCycles().size() * (series.getCycles().size() + 1.0))) + "%"); index++;
				
				if (corr(series.getCycles().get(i), series.getCycles().get(j)) >= series.getCorrelation())
					links.add(new Link(nodes.get(series.getCycles().get(i)), nodes.get(series.getCycles().get(j))));
			}
		
		print(TAB + "100%" + TAB);
		
		return new CycleNetwork(new ArrayList<Node>(nodes.values()), links, series.isMaximum(), series.getCorrelation());
	}
	
	static CorrelationNetwork correlation(Series series) {
		print("\n\t0%");
		
		long gap = (long)Math.ceil(GAP * series.getSegments().size() * (series.getSegments().size() + 1.0) / 2.0); int index = 0;
		
		Map<Segment, Node> nodes = new HashMap<Segment, Node>();
		for (Segment segment: series.getSegments()) {
			if (index > 0 && index % gap == 0) print(TAB + Math.round(200.0 * index / (series.getSegments().size() * (series.getSegments().size() + 1.0))) + "%"); index++;
			
			nodes.put(segment, new Node(segment.getId(), segment.toString()));
		}
		
		List<Link> links = new ArrayList<Link>();
		for (int i = 0; i < series.getSegments().size(); i++)
			for (int j = i + 1; j < series.getSegments().size(); j++) {
				if (index > 0 && index % gap == 0) print(TAB + Math.round(200.0 * index / (series.getSegments().size() * (series.getSegments().size() + 1.0))) + "%"); index++;
				
				if (corr(series.getSegments().get(i), series.getSegments().get(j)) >= series.getCorrelation())
					links.add(new Link(nodes.get(series.getSegments().get(i)), nodes.get(series.getSegments().get(j))));
			}
		
		print(TAB + "100%" + TAB);
		
		return new CorrelationNetwork(new ArrayList<Node>(nodes.values()), links, series.getSegments().get(0).size(), series.getCorrelation());
	}
	
	static VisibilityNetwork visibility(Series series) {
		print("\n\t0%");
		
		long gap = (long)Math.ceil(GAP * series.size() * (series.size() + 1.0) / 2.0); int index = 0;
		
		Map<Instance, Node> nodes = new HashMap<Instance, Node>();
		for (Instance instance: series) {
			if (index > 0 && index % gap == 0) print(TAB + Math.round(200.0 * index / (series.size() * (series.size() + 1.0))) + "%"); index++;
			
			nodes.put(instance, new Node(instance.getId(), instance.toString()));
		}
		
		List<Link> links = new ArrayList<Link>();
		for (int i = 0; i < series.size(); i++)
			for (int j = i + 1; j < series.size(); j++) {
				if (index > 0 && index % gap == 0) print(TAB + Math.round(200.0 * index / (series.size() * (series.size() + 1.0))) + "%"); index++;
				
				boolean visible = true;
				if (i + 1 < j) {
					double ratio = (series.get(i).getValue() - series.get(j).getValue()) / (series.get(j).getTimestamp().getTime() - series.get(i).getTimestamp().getTime());
					for (int k = i + 1; k < j && visible; k++)
						if (series.get(k).getValue() >= series.get(j).getValue() + ratio * (series.get(j).getTimestamp().getTime() - series.get(k).getTimestamp().getTime()))
							visible = false;
				}
				
				if (visible)
					links.add(new Link(nodes.get(series.get(i)), nodes.get(series.get(j))));
			}
		
		print(TAB + "100%" + TAB);
		
		return new VisibilityNetwork(new ArrayList<Node>(nodes.values()), links);
	}
	
	static TransitionNetwork transition(Series series) {
		print("\n\t0%");
		
		long gap = (long)Math.ceil(GAP * (series.getQuantiles().size() + series.size() - 1.0)); int index = 0;
		
		Map<Quantile, Node> nodes = new HashMap<Quantile, Node>();
		for (Quantile quantile: series.getQuantiles()) {
			if (index > 0 && index % gap == 0) print(TAB + Math.round(100.0 * index / (series.getQuantiles().size() + series.size() - 1.0)) + "%"); index++;
			
			nodes.put(quantile, new Node(quantile.getId(), quantile.toString()));
		}
		
		List<Link> links = new ArrayList<Link>();
		for (int i = 0; i < series.size() - 1; i++) {
			if (index > 0 && index % gap == 0) print(TAB + Math.round(100.0 * index / (series.getQuantiles().size() + series.size() - 1.0)) + "%"); index++;
			
			links.add(new Link(nodes.get(series.get(i).getQuantile()), nodes.get(series.get(i + 1).getQuantile())));
		}
		
		print(TAB + "100%" + TAB);
			
		return new TransitionNetwork(new ArrayList<Node>(nodes.values()), links, series.getQuantiles().size());
	}
	
	/*
	 * 		I/O
	 */
	
	static void json(String path, String file, int instances) throws IOException {
		print("\n\t0%");
		
		BufferedWriter writer = new BufferedWriter(new FileWriter(path + "/" + file + ".json"));
		
		long gap = (long)Math.ceil(GAP * instances);
		
		long stamp = 1000000000000000L;
		Random random = new Random(2014);
		for (int i = 0; i < instances; i++) {
			if (i > 0 && i % gap == 0) print(TAB + Math.round(100.0 * i / instances) + "%");
			
			JSONObject JSON = new JSONObject();
			
			stamp += Math.max(1000L, 25000000L + random.nextGaussian() * 5000000L);
			double value = Math.max(0.0, 1000.0 + random.nextGaussian() * 10.0);
			int trades = (int)Math.max(0, Math.round(10.0 + random.nextGaussian() * 10.0));
			double volume = Math.max(0.0, 10.0 + random.nextGaussian() * 10.0);
			
			try {
				JSON.put("_id", stamp);
			} catch (JSONException e) {
				e.printStackTrace();
			}
			
			try {
				JSON.put("avg", value);
			} catch (JSONException e) {
				e.printStackTrace();
			}
			
			try {
				JSON.put("trades", trades);
			} catch (JSONException e) {
				e.printStackTrace();
			}
			
			try {
				JSON.put("volume", volume);
			} catch (JSONException e) {
				e.printStackTrace();
			}
			
			writer.write(JSON + "\n");
		}
		
		writer.flush(); 
		writer.close();
		
		print(TAB + "100%" + TAB);
	}
	
	static Series json(String path, String file, boolean maximum, int segment, int quantile, double correlation) throws IOException {
		print("\n\t0%");
		
		List<Instance> instances = new ArrayList<Instance>();
		
		BufferedReader reader = new BufferedReader(new FileReader(path + "/" + file + ".json"));
		
		int i = 0, size = 0; String line; 
		while ((line = reader.readLine()) != null)
			if (line.trim().length() > 0)
				size++;
		
		reader.close();
		
		long gap = (long)Math.ceil(GAP * size);
		
		reader = new BufferedReader(new FileReader(path + "/" + file + ".json"));
		while ((line = reader.readLine()) != null)
			if (line.trim().length() > 0) {
				if (i > 0 && i % gap == 0) print(TAB + Math.round(100.0 * i / size) + "%"); i++;
				
				JSONObject JSON;
				try {
					JSON = new JSONObject(line);
				} catch (JSONException e) {
					e.printStackTrace();

					continue;
				}

				Calendar timestamp = Calendar.getInstance();
				try {
					if (JSON.has("_id"))
						timestamp.setTimeInMillis(JSON.getLong("_id"));
				} catch (JSONException e) {
					e.printStackTrace();
				}

				double value = -1.0;
				try {
					if (JSON.has("avg"))
						value = JSON.getDouble("avg");
				} catch (JSONException e) {
					e.printStackTrace();
				}

				int trades = -1;
				try {
					if (JSON.has("trades"))
						trades = JSON.getInt("trades");
				} catch (JSONException e) {
					e.printStackTrace();
				}
				
				double volume = -1.0;
				try {
					if (JSON.has("volume"))
						volume = JSON.getDouble("volume");
				} catch (JSONException e) {
					e.printStackTrace();
				}

				instances.add(new Instance(instances.size(), timestamp.getTime(), value, trades, volume));
			}
		
		reader.close();
		
		print(TAB + "100%" + TAB);

		return new Series(instances, maximum, segment, quantile, correlation);
	}
	
	static void arff(String path, String file, Series series) throws IOException {
		print("\n\t0%");
		
		BufferedWriter writer = new BufferedWriter(new FileWriter(path + "/" + file + ".arff"));

		writer.write("% " + BITNETS + " - The \"" + file + "\" time series in ARFF format\n%\n");
		writer.write("% The time series in STREAMS attribute-relation file format.\n%\n");
		writer.write(signature("http://www.cs.waikato.ac.nz/~ml/weka/arff.html", "%"));
		
		writer.write("\n@relation " + file + "\n");
		writer.write("@attribute id string\n");
		writer.write("@attribute bitstamp string\n");
		writer.write("@attribute timestamp date \"d/M/yy H:mm:ss\"\n");
		writer.write("@attribute trades numeric\n");
		writer.write("@attribute volume numeric\n");
		writer.write("@attribute value numeric\n"); 
		writer.write("\n@data\n");
		
		long gap = (long)Math.ceil(GAP * series.size());
		
		for (int i = 0; i < series.size(); i++) {
			if (i > 0 && i % gap == 0) print(TAB + Math.round(100.0 * i / series.size()) + "%");
			
			writer.write(String.format("%d,%d,\"%te/%tm/%ty %tk:%tM:%tS\",%d,%f,%f\n", series.get(i).getId(), series.get(i).getTimestamp().getTime(), series.get(i).getTimestamp(), series.get(i).getTimestamp(), series.get(i).getTimestamp(), series.get(i).getTimestamp(), series.get(i).getTimestamp(), series.get(i).getTimestamp(), series.get(i).getTrades(), series.get(i).getVolume(), series.get(i).getValue()));
		}
		
		writer.flush(); 
		writer.close();
		
		print(TAB + "100%" + TAB);
	}
	
	static void tab(String path, String file, Series series) throws IOException {
		print("\n\t0%");
		
		BufferedWriter writer = new BufferedWriter(new FileWriter(path + "/" + file + ".tab"));

		writer.write("% " + BITNETS + " - The \"" + file + "\" time series in TAB format\n%\n");
		writer.write("% The time series in Orange tab-delimited file format.\n%\n");
		writer.write(signature("http://orange.biolab.si/docs/latest/", "%"));
		
		writer.write("\nid\tstamp\ttrades\tvolume\tvalue\n");
		writer.write("d\tc\tc\tc\tc\n");
		writer.write("meta\t\t\t\tclass\n\n");
		
		long gap = (long)Math.ceil(GAP * series.size());
		
		for (int i = 0; i < series.size(); i++) {
			if (i > 0 && i % gap == 0) print(TAB + Math.round(100.0 * i / series.size()) + "%");
			
			writer.write(String.format("%d\t%d\t%d\t%f\t%f\n", series.get(i).getId(), series.get(i).getTimestamp().getTime(), series.get(i).getTrades(), series.get(i).getVolume(), series.get(i).getValue()));
		}
		
		writer.flush(); 
		writer.close();
		
		print(TAB + "100%" + TAB);
	}
	
	static void bitnet(String path, String file, Network network) throws IOException {
		print("\n\t0%");
		
		BufferedWriter writer = new BufferedWriter(new FileWriter(path + "/" + file + "." + network.format()));
		
		writer.write("# " + BITNETS + " - The " + network.name() + " for \"" + file + "\" time series\n#\n");
		String description = network.description(); int index = description.length() / 2;
		while (index < description.length() - 1 && description.charAt(index) != ' ') index++;
		if (index < description.length() - 1)
			writer.write("# " + description.substring(0, index) + "\n# " + description.substring(index + 1) + "\n#\n");
		else
			writer.write("# " + description + "\n#\n");
		writer.write(signature(network.reference(), "#"));
		
		long gap = (long)Math.ceil(GAP * (network.getNodes().size() + network.getLinks().size())); index = 0;
		
		writer.write("# \n# *vertices " + network.getNodes().size() + "\n");
		for (Node node: network.getNodes()) {
			if (index > 0 && index % gap == 0) print(TAB + Math.round(100.0 * index / (network.getNodes().size() + network.getLinks().size())) + "%"); index++;
			
			writer.write("# " + node + "\n");
		}
			
		writer.write("# \n# *" + (network.isDirected()? "arcs": "edges") + " " + network.getLinks().size() + "\n");
		for (Link link: network.getLinks()) {
			if (index > 0 && index % gap == 0) print(TAB + Math.round(100.0 * index / (network.getNodes().size() + network.getLinks().size())) + "%"); index++;
			
			writer.write(link + "\n");
		}
		
		writer.flush(); 
		writer.close();
		
		print(TAB + "100%" + TAB);
	}

	/*
	 * 		AUXILIARY
	 */
	
	static double ave(Segment segment) {
		return ave(segment, 0, segment.size());
	}
	
	static double ave(Segment segment, int index, int size) {
		double sum = 0.0;
		for (int i = index; i < index + size; i++)
			sum += segment.get(i).getValue();
		return sum / size;
	}
	
	static double std(Segment segment, double ave) {
		return std(segment, 0, segment.size(), ave);
	}
	
	static double std(Segment segment, int index, int size, double ave) {
		double std = 0.0;
		for (int i = index; i < index + size; i++)
			std += (segment.get(i).getValue() - ave) * (segment.get(i).getValue() - ave);
		return Math.sqrt(std / (size > 1? size - 1: 1));
	}
	
	static double corr(Segment fstSegment, Segment sndSegment, int index) {
		double fstAve = ave(fstSegment); double sndAve = ave(sndSegment, index, fstSegment.size());
		double fstStd = std(fstSegment, fstAve); double sndStd = std(sndSegment, index, fstSegment.size(), sndAve);
		
		double corr = 0.0;
		for (int i = 0; i < fstSegment.size(); i++)
			corr += (fstSegment.get(i).getValue() - fstAve) * (sndSegment.get(index + i).getValue() - sndAve);
		
		return Math.abs(corr / (fstStd * sndStd * (fstSegment.size() > 1? fstSegment.size() - 1: 1)));
	}
	
	static double corr(Segment fstSegment, Segment sndSegment) {
		if (fstSegment.size() != sndSegment.size())
			return Double.NaN;
		
		return corr(fstSegment, sndSegment, 0);
	}
	
	static double corr(Cycle fstCycle, Cycle sndCycle) {
		if (fstCycle.size() > sndCycle.size())
			return corr(sndCycle, fstCycle);
		
		double corr = Double.MIN_VALUE;
		for (int i = 0; i <= sndCycle.size() - fstCycle.size(); i++)
			corr = Math.max(corr, corr(fstCycle, sndCycle, i));
		
		return corr;
	}
	
	static String signature(String reference, String comment) {
		return comment + " For details see \"" + reference + "\".\n" + comment + "\n" + 
			comment + " by " + AUTHOR + "\n" + comment + "\n" + comment + " " + Calendar.getInstance().getTime() + "\n" + comment + "\n";	
	}
	
	static void time(long tic) {
		long millis = System.currentTimeMillis() - tic; long seconds = millis / 1000L; long minutes = seconds / 60L;
		print(BITNETS + " - " + (minutes > 0L? minutes + " min ": "") + (seconds - 60L * minutes) + " s " + (millis - 1000L * seconds) + " ms\n");
	}
	
	static void print(String print) {
		if (PRINT) 
			System.out.print(print);
	}
	
	static void error() {
		print("Error!\n");
	}
	
	static void ok() {
		print("Ok.\n");
	}
	
}

/*
 * 		CLASSES
 */

class Series extends ArrayList<Instance> {
	
	private static final long serialVersionUID = 1L;
	
	private boolean maximum;
	
	private List<Cycle> cycles;

	private List<Segment> segments;
	
	private List<Quantile> quantiles;
	
	private double correlation;
	
	public Series(List<Instance> instances, boolean maximum, int segment, int quantile, double correlation) {
		super(instances);
		this.maximum = maximum;
		this.correlation = correlation;
		
		segment = Math.max(1, Math.min(segment, size() - 1));
		quantile = Math.max(1, Math.min(quantile, size() - 1));
		
		Collections.sort(this, new Comparator<Instance>() {
			@Override
			public int compare(Instance fst, Instance snd) {
				if (fst.getValue() == snd.getValue())
					return fst.compareTo(snd);
				
				return new Double(fst.getValue()).compareTo(snd.getValue());
			}
		});
		
		quantiles = new ArrayList<Quantile>();
		for (int i = 0; i < quantile; i++)
			quantiles.add(new Quantile(i));
		
		for (int i = 0; i < size(); i++) {
			get(i).setQuantile(quantiles.get((int)(1.0 * i / size() * quantile)));
			get(i).getQuantile().include(get(i));
		}

		Collections.sort(this);
		
		cycles = new ArrayList<Cycle>(); cycles.add(new Cycle(0));
		for (int i = 0; i < size(); i++) {
			cycles.get(cycles.size() - 1).include(get(i));
			get(i).setCycle(cycles.get(cycles.size() - 1));
			
			if (i > 0 && i < size() - 1 && (maximum && get(i).getValue() > Math.max(get(i - 1).getValue(), get(i + 1).getValue()) || !maximum && get(i).getValue() < Math.min(get(i - 1).getValue(), get(i + 1).getValue()))) 
				cycles.add(new Cycle(cycles.size()));
		}
		
		segments = new ArrayList<Segment>();
		for (int i = 0; i < size() - segment + 1; i++)
			segments.add(new Segment(i));
		
		for (int i = 0; i < segments.size(); i++) {
			get(i).setSegment(segments.get(i));
			
			for (int j = 0; j < segment; j++)
				get(i).getSegment().include(get(i + j));	
		}
	}
	
	public boolean isMaximum() {
		return maximum;
	}
	
	public List<Cycle> getCycles() {
		return cycles;
	}
	
	public List<Segment> getSegments() {
		return segments;
	}

	public List<Quantile> getQuantiles() {
		return quantiles;
	}
	
	public double getCorrelation() {
		return correlation;
	}
	
}

class Segment extends ArrayList<Instance> implements Comparable<Segment> {
	
	private static final long serialVersionUID = 1L;
	
	private int id;

	public Segment(int id) {
		super();
		this.id = id;
	}
	
	public void include(Instance instance) {
		add(instance);
	}

	public int getId() {
		return id;
	}
	
	@Override
	public int compareTo(Segment segment) {
		if (id == segment.getId() && size() > 0 && segment.size() > 0)
			if (get(0).equals(segment.get(0)))
				return get(size() - 1).compareTo(segment.get(segment.size() - 1));
			else
				return get(0).compareTo(segment.get(0));
			
		return new Integer(id).compareTo(segment.getId());
	}
	
	@Override
	public String toString() {
		return String.format("{id:%d,count:%d,stamps:%d-%d,span:%d}", id, size(), size() > 0? get(0).getTimestamp().getTime(): 0L, size() > 0? get(size() - 1).getTimestamp().getTime(): 0L, size() > 0? get(size() - 1).getTimestamp().getTime() - get(0).getTimestamp().getTime(): 0L);
	}
	
}

class Cycle extends Segment {
	
	private static final long serialVersionUID = 1L;
	
	public Cycle(int id) {
		super(id);
	}
	
}

class Quantile implements Comparable<Quantile> {

	private int id;
	
	private double minimum;
	
	private double maximum;
	
	private int count;
	
	public Quantile(int id) {
		super();
		this.id = id;
		
		minimum = Double.MAX_VALUE;
		maximum = -Double.MAX_VALUE;
		count = 0;
	}
	
	public void include(Instance instance) {
		if (instance.getValue() < minimum)
			minimum = instance.getValue();
		if (instance.getValue() > maximum)
			maximum = instance.getValue();
		
		count++;
	}
	
	public int getId() {
		return id;
	}

	public double getMinimum() {
		return minimum;
	}

	public double getMaximum() {
		return maximum;
	}
	
	public int getCount() {
		return count;
	}
	
	@Override
	public int compareTo(Quantile quantile) {
		if (id == quantile.getId())
			if (minimum == quantile.getMinimum())
				return new Double(maximum).compareTo(quantile.getMaximum());
			else
				return new Double(minimum).compareTo(quantile.getMinimum());
		
		return new Integer(id).compareTo(quantile.getId());
	}
	
	@Override
	public String toString() {
		return String.format("{id:%d,count:%d,values:%f-%f,span:%f}", id, count, minimum, maximum, maximum - minimum);
	}
	
}

class Instance implements Comparable<Instance> {
	
	private int id;
	
	private Date timestamp;
	
	private double value;
	
	private int trades;
	
	private double volume;
	
	private Cycle cycle;
	
	private Segment segment;
	
	private Quantile quantile;

	public Instance(int id, Date timestamp, double value, int trades, double volume) {
		super();
		this.id = id;
		this.timestamp = timestamp;
		this.value = value;
		this.trades = trades;
		this.volume = volume;
		
		cycle = null;
		segment = null;
		quantile = null;
	}
	
	public int getId() {
		return id;
	}

	public Date getTimestamp() {
		return timestamp;
	}

	public double getValue() {
		return value;
	}

	public int getTrades() {
		return trades;
	}
	
	public double getVolume() {
		return volume;
	}
	
	public Cycle getCycle() {
		return cycle;
	}

	public void setCycle(Cycle cycle) {
		this.cycle = cycle;
	}

	public Segment getSegment() {
		return segment;
	}

	public void setSegment(Segment segment) {
		this.segment = segment;
	}

	public Quantile getQuantile() {
		return quantile;
	}

	public void setQuantile(Quantile quantile) {
		this.quantile = quantile;
	}

	@Override
	public int compareTo(Instance instance) {
		if (timestamp.equals(instance.getTimestamp()))
			return new Integer(id).compareTo(instance.getId());
		
		return timestamp.compareTo(instance.getTimestamp());
	}
	
	@Override
	public String toString() {
		return String.format("{id:%d,stamp:%d,value:%f,trades:%d,volume:%f}", id, timestamp.getTime(), value, trades, volume);
	}
	
}

enum Type {
	
	CYCLE,
	CORRELATION, 
	VISIBILITY, 
	TRANSITION;

	@Override
	public String toString() {
		switch (this) {
		case CYCLE:
			return "cycle network";
		case CORRELATION:
			return "correlation network";
		case VISIBILITY:
			return "visibility graph";
		case TRANSITION:
			return "transition network";
		default:
			break;
		}
		
		return super.toString();
	}

}

abstract class Network {
	
	private Type type;
	
	private List<Node> nodes;
	
	private List<Link> links;
	
	private boolean directed;
	
	public Network(Type type, List<Node> nodes, List<Link> links) {
		this(type, nodes, links, false);
	}

	public Network(Type type, List<Node> nodes, List<Link> links, boolean directed) {
		super();
		this.type = type;
		this.nodes = nodes;
		this.links = links;
		this.directed = directed;
		
		Collections.sort(nodes);
		/*Collections.sort(links);*/
	}
	
	public String name() {
		return type.toString();
	}
	
	public abstract String description();
	
	public abstract String reference();
	
	public abstract String format();
	
	public Type getType() {
		return type;
	}

	public List<Node> getNodes() {
		return nodes;
	}

	public List<Link> getLinks() {
		return links;
	}
	
	public boolean isDirected() {
		return directed;
	}
	
}

class CycleNetwork extends Network {
	
	private boolean maximum;
	
	private double correlation;
	
	public CycleNetwork(List<Node> nodes, List<Link> links, boolean maximum, double correlation) {
		super(Type.CYCLE, nodes, links, false);
		this.maximum = maximum;
		this.correlation = correlation;
	}
	
	@Override
	public String format() {
		return "cycnet";
	}

	@Override
	public String description() {
		return "Nodes are nonoverlapping cycles between local " + (maximum? "maximum": "minimum") + " instances of the time series and links are standard correlation coefficients between the cycles that >= " + correlation + " (simple undirected graph).";
	}

	@Override
	public String reference() {
		return "Zhang & Small, Phys. Rev. Lett. 96(23), 238701 (2006).";
	}

	public boolean isMaximum() {
		return maximum;
	}

	public double getCorrelation() {
		return correlation;
	}
	
}

class CorrelationNetwork extends Network {
	
	private int segment;
	
	private double correlation;
	
	public CorrelationNetwork(List<Node> nodes, List<Link> links, int segment, double correlation) {
		super(Type.CORRELATION, nodes, links, false);
		this.segment = segment;
		this.correlation = correlation;
	}
	
	@Override
	public String format() {
		return "cornet";
	}
	
	@Override
	public String description() {
		return "Nodes are overlapping segments of " + segment + " instances of the time series and links are standard correlation coefficients between the segments that >= " + correlation + " (simple undirected graph).";
	}

	@Override
	public String reference() {
		return "Yang & Yang, Physica A 387(5-6), 1381-1386 (2008).";
	}

	public int getSegment() {
		return segment;
	}

	public double getCorrelation() {
		return correlation;
	}
	
}

class VisibilityNetwork extends Network {
	
	public VisibilityNetwork(List<Node> nodes, List<Link> links) {
		super(Type.VISIBILITY, nodes, links, false);
	}

	@Override
	public String format() {
		return "visnet";
	}

	@Override
	public String description() {
		return "Nodes are individual instances of the time series and links are lines between instances that do not intersect any intermediate instance height (simple undirected graph).";
	}
	
	@Override
	public String reference() {
		return "Lacasa et al., PNAS 105(13), 4972-4975 (2008).";
	}
	
}

class TransitionNetwork extends Network {
	
	private int quantile;
	
	public TransitionNetwork(List<Node> nodes, List<Link> links, int quantile) {
		super(Type.TRANSITION, nodes, links, true);
		this.quantile = quantile;
	}

	@Override
	public String format() {
		return "tranet";
	}

	@Override
	public String description() {
		return "Nodes are " + quantile + "-quantiles of the time series and links are transitions between the " + quantile + "-quantiles of the consecutive instances (pseudo directed graph).";
	}
	
	@Override
	public String reference() {
		return "Campanharo et al., PLoS ONE 6(8), e23378 (2011).";
	}

	public int getQuantile() {
		return quantile;
	}
	
}

class Node implements Comparable<Node> {
	
	private int id;
	
	private String description;

	public Node(int id, String description) {
		super();
		this.id = id;
		this.description = description;
	}

	public int getId() {
		return id;
	}

	public String getDescription() {
		return description;
	}

	@Override
	public int compareTo(Node node) {
		return new Integer(id).compareTo(node.getId());
	}
	
	@Override
	public String toString() {
		return String.format("%d \"%s\"", id + 1, description);
	}
	
}

class Link implements Comparable<Link> {
	
	private Node first;
	
	private Node second;

	public Link(Node first, Node second) {
		super();
		this.first = first;
		this.second = second;
	}

	public Node getFirst() {
		return first;
	}

	public Node getSecond() {
		return second;
	}

	@Override
	public int compareTo(Link link) {
		if (first.equals(link.getFirst()))
			return second.compareTo(link.getSecond());
		
		return first.compareTo(link.getFirst());
	}
	
	@Override
	public String toString() {
		return String.format("%d %d", first.getId() + 1, second.getId() + 1);
	}
	
}
